// Copyright (C) 2024 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#include "dsmanager.h"
#include "ds.h"

#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QUdpSocket>

DesignStudioManager::DesignStudioManager(const QString &deviceUuid, QObject *parent)
    : QObject(parent)
    , m_deviceUuid(deviceUuid)
    , m_webSocketServer("DesignStudio", QWebSocketServer::NonSecureMode)
{}

void DesignStudioManager::init()
{
    static bool initialized = false;
    if (initialized) {
        return;
    }
    initialized = true;

    m_webSocketServer.listen(QHostAddress::Any, 40000);
    connect(&m_webSocketServer,
            &QWebSocketServer::newConnection,
            this,
            &DesignStudioManager::incomingConnection);
    qDebug() << "TCP server listening on port 40000";

    m_discoveryTimer.setInterval(10000);
    connect(&m_discoveryTimer, &QTimer::timeout, [this]() {
        QJsonObject message;
        message["name"] = "__qtuiviewer__";
        message["id"] = m_deviceUuid;
        QByteArray datagram = QJsonDocument(message).toJson(QJsonDocument::Compact);

        QUdpSocket udpSocket;
        const int port = 53452;

#ifdef Q_OS_ANDROID
        udpSocket.writeDatagram(datagram, QHostAddress::LocalHost, port);
#endif
        udpSocket.writeDatagram(datagram, QHostAddress::Broadcast, port);
        udpSocket.writeDatagram(datagram, QHostAddress::AnyIPv4, port);
        udpSocket.writeDatagram(datagram, QHostAddress::AnyIPv6, port);
        udpSocket.writeDatagram(datagram, QHostAddress::Any, port);
        // this is useful for Android emulator
        udpSocket.writeDatagram(datagram, QHostAddress("10.0.2.2"), port);
    });
    m_discoveryTimer.start();
}

void DesignStudioManager::incomingConnection()
{
    auto *socket = m_webSocketServer.nextPendingConnection();
    if (!socket) {
        return;
    }

    auto designStudio = QSharedPointer<DesignStudio>(new DesignStudio(socket, m_deviceUuid, this),
                                                     &QObject::deleteLater);
    connect(designStudio.data(), &DesignStudio::disconnected, this, [this, designStudio]() {
        m_designStudios.removeOne(designStudio);
        qDebug() << "Remaining Design Studios:" << m_designStudios.size();
        if (m_designStudios.isEmpty())
            emit allDesignStudiosDisconnected();
    });

    connect(designStudio.data(),
            &DesignStudio::projectIncoming,
            this,
            &DesignStudioManager::projectIncoming);

    connect(designStudio.data(),
            &DesignStudio::projectIncomingProgress,
            this,
            &DesignStudioManager::projectIncomingProgress);

    connect(designStudio.data(),
            &DesignStudio::projectVersionMismatch,
            this,
            &DesignStudioManager::projectVersionMismatch);

    connect(designStudio.data(),
            &DesignStudio::projectReceived,
            this,
            &DesignStudioManager::projectReceived);

    connect(designStudio.data(),
            &DesignStudio::projectStopRequested,
            this,
            &DesignStudioManager::projectStopRequested);

    m_designStudios.append(designStudio);
    emit designStudioConnected(designStudio->id(), designStudio->ipv4Addr());
}

void DesignStudioManager::sendProjectStarting(const QString &id)
{
    for (const auto &designStudio : m_designStudios) {
        if (designStudio->id() == id)
            designStudio->sendProjectStarting();
    }
}

void DesignStudioManager::sendProjectRunning(const QString &id)
{
    for (const auto &designStudio : m_designStudios) {
        if (designStudio->id() == id)
            designStudio->sendProjectRunning();
    }
}

void DesignStudioManager::sendProjectStopped(const QString &id)
{
    for (const auto &designStudio : m_designStudios) {
        if (designStudio->id() == id)
            designStudio->sendProjectStopped();
    }
}

void DesignStudioManager::sendProjectLogs(const QString &id, const QString &logs)
{
    for (const auto &designStudio : m_designStudios) {
        if (designStudio->id() == id)
            designStudio->sendProjectLogs(logs);
    }
}

QString DesignStudioManager::getDesignStudioIp(const QString &id) const
{
    for (const auto &designStudio : m_designStudios) {
        if (designStudio->id() == id)
            return designStudio->ipv4Addr();
    }

    return {};
}

bool DesignStudioManager::disconnectDesignStudio(const QString &id)
{
    for (const auto &designStudio : m_designStudios) {
        if (designStudio->id() == id) {
            qDebug() << "Disconnecting Design Studio" << id;
            designStudio->disconnect();
            return true;
        }
    }

    qWarning() << "Design Studio" << id << "not found for disconnection";
    return false;
}

void DesignStudioManager::disconnectAllDesignStudios()
{
    qDebug() << "Disconnecting all Design Studios";
    for (const auto &designStudio : m_designStudios)
        designStudio->disconnect();
}
