// Copyright (C) 2024 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#include "qrscanner.h"

#include <QApplication>
#include <QCamera>
#include <QCameraDevice>
#include <QImageCapture>
#include <QMediaDevices>
#include <QMessageBox>
#include <QPermission>
#include <QVideoSink>
#include <QtConcurrent>

#include "../3rdparty/zxing-cpp/example/ZXingQtReader.h"

using namespace ZXingQt;

QrScanner::~QrScanner()
{
    if (m_captureSession) {
        m_captureSession->camera()->stop();
        m_captureSession->camera()->deleteLater();
        m_captureSession->videoOutput()->deleteLater();
        m_captureSession.clear();
    }
}

void QrScanner::scanQrCode()
{
    // request permissions
    QCameraPermission permission;
    QCoreApplication &app = *QCoreApplication::instance();
    switch (app.checkPermission(permission)) {
    case Qt::PermissionStatus::Granted:
        openCamera();
        break;
    case Qt::PermissionStatus::Undetermined:
    case Qt::PermissionStatus::Denied:
        app.requestPermission(permission, [this](const QPermission &permission) {
            if (permission.status() == Qt::PermissionStatus::Denied) {
                QMessageBox msgBox{QMessageBox::Critical,
                                   "Critical:",
                                   "Camera permission denied",
                                   QMessageBox::Ok};
                msgBox.exec();
                return;
            }

            openCamera();
        });
        break;
    }
}

void QrScanner::openCamera()
{
    // start camera
    m_captureSession.reset(new QMediaCaptureSession(this));
    m_captureSession->setCamera(new QCamera(this));
    m_captureSession->setVideoOutput(new CustomVideoWidget);
    m_captureSession->camera()->setFocusMode(QCamera::FocusModeAuto);

    m_videoWidget = qobject_cast<CustomVideoWidget *>(m_captureSession->videoOutput());
    m_processing = 0;
    ZXing::ReaderOptions readerOptions;
    readerOptions.setTryRotate(false);
    readerOptions.setTryHarder(false);
    readerOptions.setTryInvert(false);
    readerOptions.setDownscaleFactor(2);

    connect(m_captureSession->videoSink(),
            &QVideoSink::videoFrameChanged,
            this,
            [&, readerOptions](const QVideoFrame &frame) {
                if (!frame.isValid())
                    return;

                if (!m_processing.testAndSetOrdered(0, 1)) {
                    return; // Another task is already running
                }

                QtConcurrent::run([=] {
                    if (!m_captureSession)
                        return QList<Result>();

                    QElapsedTimer timer;
                    timer.start();
                    QList<Result> results = ReadBarcodes(frame.toImage(), readerOptions);
                    qDebug() << "Barcode detection took" << timer.elapsed() << "ms";

                    if (!results.size())
                        m_processing = 0;

                    return results;
                }).then([=](const QList<Result> &results) {
                    if (results.isEmpty() || !m_captureSession)
                        return;

                    Result result = results.first();

                    qDebug() << "Text:   " << result.text();
                    qDebug() << "Format: " << result.format();
                    qDebug() << "Content:" << result.contentType();
                    qDebug() << "Position:" << result.position().bottomLeft()
                             << result.position().bottomRight() << result.position().topLeft()
                             << result.position().topRight();

                    emit qrCodeScanned(result.text());
                });
            });

    // stop camera when app is not active.
    // i.e. when the user switches to another app or the screen is locked
    QGuiApplication *app(qobject_cast<QGuiApplication *>(QCoreApplication::instance()));
    connect(app, &QGuiApplication::applicationStateChanged, this, [&](Qt::ApplicationState state) {
        if (state != Qt::ApplicationState::ApplicationActive && m_captureSession) {
            qDebug() << "Application is not active. Stopping camera";
            qobject_cast<CustomVideoWidget *>(m_captureSession->videoOutput())->close();
        }
    });

    // stop camera when video widget is closed
    connect(m_videoWidget, &CustomVideoWidget::closed, this, [&] {
        qDebug() << "Video widget closed.";
        emit windowClosed();
    });

    m_videoWidget->setWindowFlags(Qt::Window | Qt::WindowStaysOnTopHint);
    m_videoWidget->show();
    m_captureSession->camera()->start();
}
