// Copyright (C) 2024 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#include "settings.h"

#include <QDir>
#include <QFile>
#include <QJsonDocument>
#include <QStandardPaths>

/*
    Example of a settings.json file:
    {
        "autoScale": true
        "deviceUuid": "1234567890"
    }
*/

Settings::Settings()
    : m_settingsPath(
        QStandardPaths::writableLocation(QStandardPaths::ConfigLocation).append("/settings.json"))
{
    qDebug() << "Settings path:" << m_settingsPath;
    const QDir dir(QStandardPaths::writableLocation(QStandardPaths::ConfigLocation));

    if (!dir.exists()) {
        qWarning() << "Settings directory does not exist. Creating one.";
        dir.mkpath(QStandardPaths::writableLocation(QStandardPaths::ConfigLocation));
    }

    if (!loadSettings()) {
        qDebug() << "Failed to load settings. Applying default ones.";
        applyDefaultSettings();
    }
}

void Settings::applyDefaultSettings()
{
    m_settings["autoScale"] = true;
    m_settings["deviceUuid"] = "";
    saveSettings();
}

bool Settings::loadSettings()
{
    QFile file(m_settingsPath);
    if (!file.open(QIODevice::ReadOnly)) {
        qWarning() << "Failed to open settings file" << m_settingsPath
                   << "Reason:" << file.errorString();
        return false;
    }

    QJsonDocument doc = QJsonDocument::fromJson(file.readAll());
    m_settings = doc.object();
    return true;
}

bool Settings::saveSettings()
{
    QFile file(m_settingsPath);
    if (!file.open(QIODevice::WriteOnly)) {
        qWarning() << "Failed to save settings to" << m_settingsPath
                   << "Reason:" << file.errorString();
        return false;
    }

    QJsonDocument doc(m_settings);
    file.write(doc.toJson());
    return true;
}

void Settings::setAutoScaleProject(const bool &enabled)
{
    m_settings["autoScale"] = enabled;
    saveSettings();
}

void Settings::setDeviceUuid(const QString &deviceUuid)
{
    m_settings["deviceUuid"] = deviceUuid;
    saveSettings();
}

void Settings::setKeepScreenOn(const bool &enabled)
{
    m_settings["keepScreenOn"] = enabled;
    saveSettings();
}

QString Settings::deviceUuid() const
{
    return m_settings["deviceUuid"].toString();
}

bool Settings::autoScaleProject() const
{
    return m_settings["autoScale"].toBool();
}

bool Settings::keepScreenOn() const
{
    return m_settings["keepScreenOn"].toBool();
}
