// Copyright (C) 2024 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#include <QGuiApplication>
#include <QTemporaryFile>
#include <QTest>

#include "tst_settings.h"

#define DEBUG_LINE qDebug() << "Debugline:" << __LINE__

int main(int argc, char *argv[])
{
    QGuiApplication app(argc, argv);

    QStringList initialArgs = app.arguments();
    DEBUG_LINE << "Arguments:" << initialArgs;

    QString outputFileName;
    QString appName = initialArgs.first();
    initialArgs.removeFirst();

    if (initialArgs.contains("-o")) {
        int index = initialArgs.indexOf("-o");
        if (index < initialArgs.size() - 1) {
            outputFileName = initialArgs.at(index + 1);
            outputFileName.remove(",junitxml"); // remove the junitxml extension if it's provided
            DEBUG_LINE << "Find output file name inside the args:" << outputFileName;

            initialArgs.removeAt(index); // remove the -o flag
            initialArgs.removeAt(index); // remove the output file name
        }
    }

    if (outputFileName.isEmpty()) {
        qDebug() << "Output file name not provided. Using the default name: output.junitxml";
        outputFileName = "output.junitxml";
    }

    QFile outputFile{outputFileName};
    if (!outputFile.open(QIODevice::Append)) {
        DEBUG_LINE << "Failed to open the output file";
        return 1;
    }

    int status = 0;
    auto runTest = [&status, &outputFile, initialArgs, appName](QObject *obj) {
        // temporary file to store the individual test results.
        // we'll append the results to the output file.
        QTemporaryFile file;
        file.open();
        DEBUG_LINE << "Temporary file:" << file.fileName();

        QStringList args{"-o", QString(file.fileName()).append(",junitxml")};
        args.prepend(appName);
        args.append(initialArgs);

        DEBUG_LINE << "Running test" << obj->metaObject()->className() << "with arguments:" << args;
        status |= QTest::qExec(obj, args);

        // append the test results to the output file
        file.readLine(); // skip the first line because it's the xml header
        outputFile.write(file.readAll());
    };

    outputFile.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
    outputFile.write("<testsuites>\n");

    runTest(new TestSettings);

    outputFile.write("</testsuites>\n");

    DEBUG_LINE << "Test suite finished with status:" << status;
    return status;
}
